[connection]
id=static
uuid={{uuid}}
type=ethernet
interface-name={{wired_iface}}
permissions=

[ethernet]
mac-address-blacklist=

[ipv4]
address1={{wired_ip}}/24
dns=8.8.8.8;
dns-search=
may-fail=false
method=manual

[ipv6]
addr-gen-mode=stable-privacy
dns-search=
method=disabled
