[connection]
id=dhcp
uuid={{uuid}}
type=ethernet
interface-name={{wired_iface}}
permissions=
autoconnect-priority=1
autoconnect-retries=4

[ethernet]

[ipv4]
method=auto
dhcp-timeout=30
dhcp-vendor-class-identifier=NM:ConsolePi


[ipv6]
addr-gen-mode=stable-privacy
method={{wired_v6_method|default('auto')}}
