[connection]
id=hotspot
uuid=b226425f-80df-489a-8013-767f31ad42f7
type=wifi
interface-name=wlan0
autoconnect=true
autoconnect-priority=1

[wifi]
band=bg
channel=1
mode=ap
ssid={{hotspot_ssid}}

[wifi-security]
group=ccmp;
key-mgmt=wpa-psk
pairwise=ccmp;
proto=rsn;
psk={{hotspot_psk}}

[ipv4]
address1={{hotspot_ip}}/24
method=shared

[ipv6]
addr-gen-mode=default
method=disabled

[proxy]