# dnsmasq configuration created by ConsolePi installer
# modifications should not be made to this file add additional options
# in new files in this directory.
interface={{ wlan_iface|default('wlan0') }}
except-interface=lo
bogus-priv
domain-needed
dhcp-range={{wlan_dhcp_start}},{{wlan_dhcp_end}},255.255.255.0,12h
bind-interfaces
resolv-file=/run/dnsmasq/resolv.conf
dhcp-option={{ wlan_iface|default('wlan0') }},3
