[connection]
id=hotspot
uuid={{uuid}}
type=wifi
interface-name={{wlan_iface}}
autoconnect=true
autoconnect-priority=-999

[wifi]
band=bg
mode=ap
ssid={{wlan_ssid}}

[wifi-security]
key-mgmt=wpa-psk
psk={{wlan_psk}}

[ipv4]
address1={{wlan_ip}}/24
method=shared

[ipv6]
addr-gen-mode=default
method={{v6_method|default('auto')}}

[proxy]
