# dnsmasq configuration created by ConsolePi installer
# modifications can be made but the 'dhcp-option-wlan0,3' line needs to exist
# for the default-gateway on/off based on wired status function to work
interface={{ wlan_iface|default('wlan0' )}}
bogus-priv
domain-needed
dhcp-range={{ wlan_dhcp_start }},{{ wlan_dhcp_end }},255.255.255.0,12h
bind-interfaces
dhcp-option={{ wlan_iface|default('wlan0') }},3
